/*
 * AdabasSql.c --
 *
 * This file creates the tcl commands of the extension (e.g. adalogon
 * and adasql). Most of the procedures do only a little option checking
 * and then call the corrensponding procedure in AdabasApi.c (e.g.
 * the tcl command adafetch is implemented by AdaFetchCmd, which ends
 * in a call of AdabasFetch from AdabasApi.c
 *
 * Copyright (c) 1996-1997 Christian Krone. All rights reserved.
 * This program is free software; you can redistribute it and/or
 * modify it under the same terms as Tcl itself.
 * See also licence.terms
 *
 * $Id: AdabasSql.c,v 1.62 1997/01/09 19:20:20 adabas Exp $
 */

/*
 * Import interface (should move into AdabasTclInt.h someday)
 */

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <tcl.h>

#include "adabas.h"
#include "AdabasPort.h"
#include "AdabasRte.h"
#include "AdabasPacket.h"
#include "AdabasFormat.h"
#include "AdabasUtil.h"
#include "AdabasSend.h"
#include "AdabasSql.h"
#include "AdabasApi.h"
#include "AdabasLong.h"

/*
 * Prototypes for procedures referenced only in this file. Since they are
 * used in the following ProcTable, they are declared before the variable
 * declarations.
 */

static int AdaLogonCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				    int objc, Tcl_Obj *objv[]));
static int AdaLogoffCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				     int objc, Tcl_Obj *objv[]));
static int AdaOpenCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				   int objc, Tcl_Obj *objv[]));
static int AdaCloseCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				    int objc, Tcl_Obj *objv[]));
static int AdaSqlCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				  int objc, Tcl_Obj *objv[]));
static int AdaCancelCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				     int objc, Tcl_Obj *objv[]));
static int AdaFetchCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				    int objc, Tcl_Obj *objv[]));
static int AdaColsCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				   int objc, Tcl_Obj *objv[]));
static int AdaReadLongCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				       int objc, Tcl_Obj *objv[]));
static int AdaWriteLongCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
					int objc, Tcl_Obj *objv[]));
static int AdaAutoCommitCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
					 int objc, Tcl_Obj *objv[]));
static int AdaCommitCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				     int objc, Tcl_Obj *objv[]));
static int AdaRollbackCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				       int objc, Tcl_Obj *objv[]));
static int AdaUtilCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				   int objc, Tcl_Obj *objv[]));
static int AdaSpecialCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				      int objc, Tcl_Obj *objv[]));

/*
 * For every tcl command implemented by this file, there is an entry
 * in the following (null terminated) table, which combines the name
 *  of the tcl command with its command procedure.
 */

typedef struct {
  char           *name;
  Tcl_ObjCmdProc *proc;
} ObjProcTable;

static ObjProcTable adabasObjProcs[] = {
  { "adalogon",     AdaLogonCmd       },
  { "adalogoff",    AdaLogoffCmd      },
  { "adaopen",      AdaOpenCmd        },
  { "adaclose",     AdaCloseCmd       },
  { "adasql",       AdaSqlCmd         },
  { "adacancel",    AdaCancelCmd      },
  { "adafetch",     AdaFetchCmd       },
  { "adacols",      AdaColsCmd        },
  { "adaautocom",   AdaAutoCommitCmd  },
  { "adacommit",    AdaCommitCmd      },
  { "adarollback",  AdaRollbackCmd    },
  { "adareadlong",  AdaReadLongCmd    },
  { "adawritelong", AdaWriteLongCmd   },
  { "adautil",      AdaUtilCmd        },
  { "adaspecial",   AdaSpecialCmd     },
  { (char *) NULL, (Tcl_ObjCmdProc *) NULL }
};

/*
 *----------------------------------------------------------------------
 *
 * AdabasSqlInit --
 *
 *      Initializes this module by registering all the command procedures.
 *
 * Results:
 *      The return value is normally TCL_OK. If there are problems by
 *      evaluating the startup script adasql.tcl, TCL_ERROR will be returned.
 *
 * Side effects:
 *      The tcl commands of the AdabasTcl extension are defined in the given
 *      interpreter.
 *
 *----------------------------------------------------------------------
 */

int
AdabasSqlInit (interp)
     Tcl_Interp *interp;		/* Current interpreter. */
{
  ObjProcTable *pObj;			/* index pointer in proc table. */
  Tcl_DString   fileName;		/* full path name of startup script. */
  int           erg;			/* tcl result to return. */
  char         *adabasTclDir;		/* environment variable ADABASTCL_LIB. */

  /*
   * Define all the tcl commands, that are mentioned in proc table.
   */

  for (pObj = adabasObjProcs; pObj->name; pObj++) {
    Tcl_CreateObjCommand (interp, pObj->name, -1, pObj->proc,
			  (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  }

  /*
   * Evaluate the startup script '$ADABASTCL_DIR/adasql.tcl' and
   * return the result of that evaluation.
   */

  adabasTclDir = Tcl_GetVar (interp, "adabastcl_library", TCL_GLOBAL_ONLY);
  Tcl_DStringInit (&fileName);
  Tcl_DStringAppend (&fileName, adabasTclDir,  -1);
  Tcl_DStringAppend (&fileName, "/adasql.tcl", -1);
  erg = Tcl_EvalFile (interp, Tcl_DStringValue (&fileName));
  Tcl_DStringFree (&fileName);
  return erg;
}

/*
 *----------------------------------------------------------------------
 *
 * AdaOpenCmd --
 *
 *      Implements the tcl command adaopen with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK; in this case interp->results
 *      will be set to the name of the fresh created cursorHandle. If there
 *      are malformed parameters or any problems while createing the cursor,
 *      TCL_ERROR will be returned.
 *
 * Side effects:
 *      The hashtable of cursorHandle will have an additional line, which
 *      contains a structure with information about the newly created cursor.
 *
 *----------------------------------------------------------------------
 */

static int
AdaOpenCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *logonName;		/* Name of logonHandle. */
  AdabasInfo *vars;			/* logonHandle identified by 1. param. */
  CursorInfo *cursorVars;		/* the newly created cursorHandle. */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */

  /*
   * Check, if the one and only parameter is a valid logonHandle. Therefore
   * it must be found and it must be created by adalogon with -service user
   * or -service control.
   */

  if (objc != 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " logonHandle\"");
    return TCL_ERROR;
  }
  logonName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (AdabasLogonHandle (interp, logonName, &vars) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if (vars->packetInfo.messType != M_DBS) {
    AppendResult (interp, "Logon handle \"");
    AppendResult (interp, logonName);
    AppendResult (interp, "\" must be user or control service");
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which open the cursor and allocates
   * some information about it.
   */

  if (!(cursorVars = AdabasOpen (interp, vars))) {
    return TCL_ERROR;
  }

  AdabasSetMsgString (interp, AdamsgHandle, cursorVars->cursorName);
  Tcl_ResetObjResult (interp);
  AppendResult (interp, cursorVars->cursorName);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * AdaCloseCmd --
 *
 *      Implements the tcl command adaclose with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid cursorName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *
 * Side effects:
 *      The hashtable of cursorHandle will have one fewer line, which used to
 *      contain a structure with information about the destroyed cursor.
 *
 *----------------------------------------------------------------------
 */

static int
AdaCloseCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *cursorName;		/* Name of cursor Handle. */
  CursorInfo *cursorVars;		/* the cursorHandle to destroy. */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */

  /*
   * Check, if the one and only parameter is a valid cursorHandle.
   */

  if (objc != 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " cursorHandle\"");
    return TCL_ERROR;
  }
  cursorName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (!(cursorVars = AdabasCursorHandle (interp, cursorName))) {
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which closes the cursor and
   * frees the information about it.
   */

  return AdabasClose (interp, cursorVars);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaSqlCmd --
 *
 *      Implements the tcl command adasql with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid cursorName
 *      or any malformed combination of options given, TCL_ERROR will be
 *      returned; in this case interp->results will be set to the error message.
 *      The same goes, if the database kernel returned a returncode <> 0.
 *
 * Side effects:
 *      Some entries in the tcl array adamsg will bet set.
 *
 *----------------------------------------------------------------------
 */

static int
AdaSqlCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *cursorName;		/* Name of cursor Handle. */
  CursorInfo *cursorVars;		/* cursorHandle given by 1. parameter. */
  int         withParameter = 0;	/* -parameter option given? */
  char       *sqlMode       = NULL;     /* value of -sqlmode option. */
  char       *resultTable   = NULL;	/* value of -resulttable option. */
  char       *parsId;			/* ParsId to execute. */
  int         thisMessType;		/* messType of this sql request. */
  char       *cmd;			/* SQL command (2. or 3. parameter). */
  char       *currOption;		/* Option string (used for parsing). */
  int         length;			/* Length (used for option parsing). */
  int         c;			/* Char   (used for option parsing). */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */
  int         missingParams;		/* Too few parameters? */

  /*
   * Check, if there are at least the cursorName and a SQL statement.
   */

  missingParams = objc < 3;
  if (!missingParams) {
    cmd = Tcl_GetStringFromObj (interp, objv[2], &length);
    missingParams = cmd[0] == '-' && objc < 4;
  }
  if (missingParams) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " cursorHandle ?kind? sqlStatement ?options?\"");
    return TCL_ERROR;
  }
  cursorName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);

  /*
   * If the second prameter starts with an hyphen, it is treated as description
   * of the messType of the SQL request. If not specified, -command is default.
   *  option      messType           comment
   *  -command    M_DBS	             pars/execute in one step (default);
   *  -parameter  M_PARSE/M_EXECUTE  parsing and executing separately;
   *  -parse      M_PARSE            only parsing, return a parsId;
   *  -execute    M_EXECUTE          only executing of the given parsId.
   */

  if (cmd[0] != '-') {
    thisMessType = M_DBS;
  } else {
    c = cmd[1];
    if (c == 'c' && !strncmp (cmd, "-command", length))
      thisMessType = M_DBS;
    else if (c == 'p' && !strncmp (cmd, "-parameter", length) && length > 4) {
      thisMessType  = M_PARSE;
      withParameter = 1;
    } else if (c == 'p' && !strncmp (cmd, "-parse", length) && length > 4)
      thisMessType = M_PARSE;
    else if (c == 'e' && !strncmp (cmd, "-execute", length)) {
      if (objc != 4) {
	AppendResult (interp, "wrong # args: should be \"");
	AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0],
						     &dummyLength));
	AppendResult (interp, " cursorHandle -execute parsId\"");
	return TCL_ERROR;
      }
      if (!(cursorVars = AdabasCursorHandle (interp, cursorName))) {
	return TCL_ERROR;
      }
      parsId = Tcl_GetStringFromObj (interp, objv[3], &dummyLength);
      return AdabasExecParsId (interp, cursorVars, parsId);

    } else if (c == 'd' && !strncmp (cmd, "-drop", length)) {

      if (objc != 4) {
	AppendResult (interp, "wrong # args: should be \"");
	AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0],
						     &dummyLength));
	AppendResult (interp, " cursorHandle -drop parsId\"");
	return TCL_ERROR;
      }
      if (!(cursorVars = AdabasCursorHandle (interp, cursorName))) {
	return TCL_ERROR;
      }
      parsId = Tcl_GetStringFromObj (interp, objv[3], &dummyLength);
      return AdabasDropParsId (interp, cursorVars, parsId);

    } else {
      AppendResult (interp, "unknown kind \"");
      AppendResult (interp, cmd);
      AppendResult (interp, "\" should be ");
      AppendResult (interp, "-command, -parameter, -parse, -execute or -drop");
      return TCL_ERROR;
    }
    objc--; objv++;
    cmd = Tcl_GetStringFromObj (interp, objv[2], &dummyLength);
  }

  /*
   * Check, if the given cursorName is the name of a valid cursorHandle,
   * and parse the trailing option switches (-resulttable or -sqlmode).
   */

  if (!(cursorVars = AdabasCursorHandle (interp, cursorName))) {
    return TCL_ERROR;
  }
  for (objv+=3, objc-=3; objc > 1; objv += 2, objc -= 2) {
    currOption = Tcl_GetStringFromObj (interp, objv[0], &length);
    c          = currOption[1];

    if (c == 'r' && !strncmp (currOption, "-resulttable", length)) {
      resultTable = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
    } else if (c == 's' && !strncmp (currOption, "-sqlmode", length)) {
      sqlMode = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
    } else {
      AppendResult (interp, "unknown or ambigious switch \"");
      AppendResult (interp, currOption);
      AppendResult (interp, "\": must be -resulttable or -sqlmode");
      return TCL_ERROR;
    }
  }
  if (objc > 0) {
    AppendResult (interp, "missing argument for switch: \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, "\"");
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which handles the SQL request.
   */

  return AdabasSql (interp, cursorVars, thisMessType, withParameter,
		    cmd, resultTable, sqlMode);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaCancelCmd --
 *
 *      Implements the tcl command adacancel with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid cursorName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *
 * Side effects:
 *      The currently active SQL request will be aborted. Cursor variables
 *      will be reset.
 *
 *----------------------------------------------------------------------
 */

static int
AdaCancelCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *cursorName;		/* Name of cursor Handle. */
  CursorInfo *cursorVars;		/* cursorHandle given by 1. parameter. */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */

  /*
   * Check, if the one and only parameter is a valid cursorHandle.
   */

  if (objc != 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " cursorHandle\"");
    return TCL_ERROR;
  }
  cursorName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (!(cursorVars = AdabasCursorHandle (interp, cursorName))) {
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which cancels the request and resets
   * all information of the cursor.
   */

  return AdabasCancel (interp, cursorVars);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaFetchCmd --
 *
 *      Implements the tcl command adafetch with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid cursorName
 *      or any malformed combination of options given, TCL_ERROR will be
 *      returned; in this case interp->results will be set to the error message.
 *      The same goes, if the database kernel returned a returncode <> (0, 100).
 *
 *      If the returnvalue is TCL_OK, interp->result will point to a valid
 *      tcl list, that is the represantation of the fetched row. If the
 *      option -count was given with a value > 1, interp->result will point
 *      to a list of lists, each represanting a fetched row.
 *
 * Side effects:
 *      The cursor is moved, so that the next positional fetch will return a
 *      different row.
 *
 *----------------------------------------------------------------------
 */

static int
AdaFetchCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *cursorName;		/* Name of cursor Handle. */
  CursorInfo *cursorVars;		/* cursorHandle given by 1. parameter. */
  char       *currOption;		/* Option string (used for parsing). */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */
  int         length;			/* Length (used for option parsing). */
  int         c;			/* Char   (used for option parsing). */
  char       *target   = "";		/* modified value of -position option. */
  long        massCnt  = 0;		/* value of -count option. */
  int         arraySet = 0;		/* value of -array option. */
  char       *sqlMode  = NULL;		/* value of -sqlmode option. */
  char       *command  = NULL;		/* value of -command option. */
  char        msg[80];			/* Used for error message. */

  /*
   * Check, if the given cursorName is the name of a valid cursorHandle,
   * and if it is currently open for a fetch (i.e. the last SQL request
   * was a successful mass select).
   */

  if (objc < 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " cursorHandle ?options?\"");
    return TCL_ERROR;
  }
  cursorName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (!(cursorVars = AdabasCursorHandle (interp, cursorName))) {
    return TCL_ERROR;
  }
  if (!cursorVars->cntParams) {
    AppendResult (interp, "\"");
    AppendResult (interp, cursorName);
    AppendResult (interp, "\" not opened for fetch");
    return TCL_ERROR;
  }
  cursorVars->dataEnd = 0;

  /*
   * Parse the trailing option switches.
   */

  for (objv+=2, objc-=2; objc > 1; objv += 2, objc -= 2) {
    currOption = Tcl_GetStringFromObj (interp, objv[0], &length);
    c          = currOption[1];

    if (c == 'a' && !strncmp (currOption, "-array", length)) {
      if (Tcl_GetBooleanFromObj (interp, objv[1], &arraySet) == TCL_ERROR) {
	return TCL_ERROR;
      }
    } else if (c == 'c' && !strncmp (currOption, "-command", length)
	       && length > 3) {
      command = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
    } else if (c == 'c' && !strncmp (currOption, "-count", length) && length > 3) {
      if (Tcl_GetIntFromObj (interp, objv[1], &massCnt) == TCL_ERROR) {
	return TCL_ERROR;
      }
      if (massCnt < 1) {
	sprintf (msg, "invalid -count %ld", massCnt);
	AppendResult (interp, msg);
	return TCL_ERROR;
      }
    } else if (c == 'p' && !strncmp (currOption, "-position", length)) {
      target = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
    } else if (c == 's' && !strncmp (currOption, "-sqlmode", length)) {
      sqlMode = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
    } else {
      AppendResult (interp, "unknown or ambigious switch \"");
      AppendResult (interp, currOption);
      AppendResult (interp, "\": must be -array, -command, -count, -position ");
      AppendResult (interp, "or -sqlmode");
      return TCL_ERROR;
    }
  }
  if (objc > 0) {
    AppendResult (interp, "missing argument for switch: \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, "\"");
    return TCL_ERROR;
  }
  if (*target && massCnt > 0) {
    SetResult (interp, "Invalid combination of -count and -position");
    return TCL_ERROR;
  }
  if (*target && command) {
    SetResult (interp, "Invalid combination of -command and -position");
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi dependend of the value of massCnt,
   * which fetches one or many rows and returns them in interp->result.
   */

  if (command) {
    if (massCnt < 1) massCnt = 20;
    return AdabasFetchAll (interp, cursorName, massCnt, command);
  }
  if (!massCnt) {
    massCnt = 1;
  }
  return AdabasFetch (interp, cursorVars, target, massCnt, sqlMode, arraySet);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaColsCmd --
 *
 *      Implements the tcl command adacols with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid cursorName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *      If the returnvalue is TCL_OK, interp->result will point to a valid
 *      tcl list, containing the columnnames of the last select of this cursor.
 *
 * Side effects:
 *      Some entries in the tcl array adamsg will bet set.
 *
 *----------------------------------------------------------------------
 */

static int
AdaColsCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *cursorName;		/* Name of cursor Handle. */
  CursorInfo *cursorVars;		/* cursorHandle given by 1. parameter. */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */

  /*
   * Check, if the one and only parameter is a valid cursorHandle.
   */

  if (objc != 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " cursorHandle\"");
    return TCL_ERROR;
  }
  cursorName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (!(cursorVars = AdabasCursorHandle (interp, cursorName))) {
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which constructs the list of columnNames.
   */

  return AdabasColumns (interp, cursorVars);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaReadLongCmd --
 *
 *      Implements the tcl command adareadlong with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid cursorName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *      If the returnvalue is TCL_OK and a -filename option was given,
 *      interp->result will point to the number of bytes written.
 *      If the returnvalue is TCL_OK and no -filename option was given,
 *      interp->result will point to the value of the long column.
 *
 * Side effects:
 *      If the returnvalue is TCL_OK and a -filename option was given,
 *      the mentioned file contains the value of the long column.
 *
 *----------------------------------------------------------------------
 */

static int
AdaReadLongCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char        *cursorName;		/* Name of cursor Handle. */
  CursorInfo  *cursorVars;		/* cursorHandle given by 1. parameter. */
  int          length;			/* Length (used for option parsing). */
  int          c;			/* Char   (used for option parsing). */
  int          dummyLength;		/* Out parameter of GetXxxFromObj. */
  char        *currOption;		/* Option string (used for parsing). */
  char        *currParameter;		/* Value  string (used for parsing). */
  char        *tableName   = NULL;	/* value of -tablename option. */
  char        *columnName  = NULL;	/* value of -columnname option. */
  char        *descrString = NULL;	/* value of -descriptor option. */
  char        *fileName    = NULL;	/* value of -filename option. */
  char        *whereCond   = NULL;	/* value of -where option. */
  DataEncoding encoding    = encNil;	/* modified value of -encoding option. */

  /*
   * Check, if the first parameter is the name of a valid cursorHandle.
   */

  if (objc <= 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " cursorHandle options...\"");
    return TCL_ERROR;
  }
  cursorName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (!(cursorVars = AdabasCursorHandle (interp, cursorName))) {
    return TCL_ERROR;
  }

  /*
   * Parse the trailing option switches.
   */
  for (objv+=2, objc-=2; objc > 1; objv += 2, objc -= 2) {
    currOption    = Tcl_GetStringFromObj (interp, objv[0], &length);
    currParameter = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
    c             = currOption[1];

    if (c == 'c' && !strncmp (currOption, "-column", length)) {
      columnName = currParameter;
    } else if (c == 'd' && !strncmp (currOption, "-descriptor", length)) {
      descrString = currParameter;
    } else if (c == 'f' && !strncmp (currOption, "-filename", length)) {
      fileName = currParameter;
    } else if (c == 'e' && !strncmp (currOption, "-encoding", length)) {
      if ((encoding = scanEncoding (interp, currParameter)) == encNil) {
	return TCL_ERROR;
      }
    } else if (c == 't' && !strncmp (currOption, "-table", length)) {
      tableName = currParameter;
    } else if (c == 'w' && !strncmp (currOption, "-where", length)) {
      whereCond = currParameter;
    } else {
      AppendResult (interp, "bad switch \"");
      AppendResult (interp, currOption);
      AppendResult (interp, "\": must be -table, -column, -where,");
      AppendResult (interp, " -descriptor, -encoding or -filename");
      return TCL_ERROR;
    }
  }
  if (objc > 0) {
    AppendResult (interp, "missing argument for switch: \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, "\"");
    return TCL_ERROR;
  }
  if ((fileName && encoding != encNil) ||
      !(( descrString && !tableName && !columnName && !whereCond) ||
	(!descrString &&  tableName &&  columnName))) {
    SetResult (interp, "invalid combination of switches.");
    return TCL_ERROR;
  }
  if (encoding == encNil) {
    encoding = encEscape;
  }

  /*
   * Call the procedure from AdabasApi, which generates the SQL requests
   * to read the long column and write it into the given file (if fileName
   * is <> NULL) or let interp->result point to it otherwise.
   */

  return AdabasReadLong (interp, cursorVars, descrString, tableName, columnName,
			 whereCond, encoding, fileName);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaWriteLongCmd --
 *
 *      Implements the tcl command adawritelong with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid cursorName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *      If the returnvalue is TCL_OK and a -filename option was given,
 *      interp->result will point to the number of bytes written.
 *      If the returnvalue is TCL_OK and no -filename option was given,
 *      interp->result will point to the value of the long column.
 *
 * Side effects:
 *      If the returnvalue is TCL_OK and a -filename option was given,
 *      the mentioned file contains the value of the long column.
 *
 *----------------------------------------------------------------------
 */

static int
AdaWriteLongCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char        *cursorName;		/* Name of cursor Handle. */
  CursorInfo  *cursorVars;		/* cursorHandle given by 1. parameter. */
  int          length;			/* Length (used for option parsing). */
  int          c;			/* Char   (used for option parsing). */
  int          dummyLength;		/* Out parameter of GetXxxFromObj. */
  char        *currOption;		/* Option string (used for parsing). */
  char        *currParameter;		/* Value  string (used for parsing). */
  char        *tableName  = NULL;	/* value of -tablename option. */
  char        *columnName = NULL;	/* value of -columnname option. */
  char        *fileName   = NULL;	/* value of -filename option. */
  char        *longValue  = NULL;	/* value of -longvalue option. */
  char        *whereCond  = NULL;	/* value of -where option. */
  DataEncoding encoding   = encNil;	/* modified value of -encoding option. */

  /*
   * Check, if the first parameter is the name of a valid cursorHandle.
   */

  if (objc <= 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " cursorHandle options...\"");
    return TCL_ERROR;
  }
  cursorName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (!(cursorVars = AdabasCursorHandle (interp, cursorName))) {
    return TCL_ERROR;
  }

  /*
   * Parse the trailing option switches.
   */

  for (objv+=2, objc-=2; objc > 1; objv += 2, objc -= 2) {
    currOption    = Tcl_GetStringFromObj (interp, objv[0], &length);
    currParameter = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
    c             = currOption[1];

    if (c == 'c' && !strncmp (currOption, "-column", length)) {
      columnName = currParameter;
    } else if (c == 'e' && !strncmp (currOption, "-encoding", length)) {
      if ((encoding = scanEncoding (interp, currParameter)) == encNil) {
	return TCL_ERROR;
      }
    } else if (c == 'f' && !strncmp (currOption, "-filename", length)) {
      fileName = currParameter;
    } else if (c == 't' && !strncmp (currOption, "-table", length)) {
      tableName = currParameter;
    } else if (c == 'v' && !strncmp (currOption, "-value", length)) {
      longValue = currParameter;
    } else if (c == 'w' && !strncmp (currOption, "-where", length)) {
      whereCond = currParameter;
    } else {
      AppendResult (interp, "bad switch \"");
      AppendResult (interp, currOption);
      AppendResult (interp, "\": must be -table, -column, -where,");
      AppendResult (interp, " -value, -encoding or -filename");
      return TCL_ERROR;
    }
  }
  if (objc > 0) {
    AppendResult (interp, "missing argument for switch: \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, "\"");
    return TCL_ERROR;
  }
  if ((fileName && encoding != encNil) || !tableName || !columnName) {
    SetResult (interp, "invalid combination of switches.");
    return TCL_ERROR;
  }
  if (!longValue && !fileName) {
    SetResult (interp, "either -filename or -value switch must be given");
    return TCL_ERROR;
  }
  if (encoding == encNil) {
    encoding = encEscape;
  }

  /*
   * Call the procedure from AdabasApi, which generates the SQL requests
   * to write the long column from the given file (if fileName
   * is <> NULL) or from the given longValue.
   */

  return AdabasWriteLong (interp, cursorVars, tableName, columnName,
			  whereCond, encoding, fileName, longValue);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaAutoCommitCmd --
 *
 *      Implements the tcl command adaautocom with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid logonName
 *      or boolean value given, TCL_ERROR will be returned; in this case
 *      interp->results will be set to the error message.
 *
 * Side effects:
 *      The autocommit flag in future SQL requests will be set accordingly.
 *
 *----------------------------------------------------------------------
 */

static int
AdaAutoCommitCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *logonName;		/* Name of logonHandle. */
  AdabasInfo *vars;			/* logonHandle identified by 1. param. */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */
  int         onOff;			/* modified value of 2. param. */

  /*
   * Check, if the first parameter is the name of a valid logonHandle
   * and if the second parameter is a valid boolean value.
   */

  if (objc != 3) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " logonHandle onOff\"");
    return TCL_ERROR;
  }
  logonName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (AdabasLogonHandle (interp, logonName, &vars) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if (Tcl_GetBooleanFromObj (interp, objv[2], &onOff) == TCL_ERROR) {
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which switches the
   * autocommit feature of the database kernel.
   */

  return AdabasAutoCommit (interp, vars, onOff);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaCommitCmd --
 *
 *      Implements the tcl command adacommit with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid logonName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *
 * Side effects:
 *      All SQL statements of the current logon are commited.
 *
 *----------------------------------------------------------------------
 */

static int
AdaCommitCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *logonName;		/* Name of logonHandle. */
  AdabasInfo *vars;			/* logonHandle identified by 1. param. */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */

  /*
   * Check, if the only parameter is the name of a valid logonHandle.
   */

  if (objc != 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " logonHandle\"");
    return TCL_ERROR;
  }
  logonName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (AdabasLogonHandle (interp, logonName, &vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which sends a COMMIT request
   * to the database kernel.
   */

  return AdabasCommit (interp, vars);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaRollbackCmd --
 *
 *      Implements the tcl command adarollback with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid logonName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *
 * Side effects:
 *      All SQL statements of the current logon are rollbacked.
 *
 *----------------------------------------------------------------------
 */

static int AdaRollbackCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *logonName;		/* Name of logonHandle. */
  AdabasInfo *vars;			/* logonHandle identified by 1. param. */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */

  /*
   * Check, if the only parameter is the name of a valid logonHandle.
   */

  if (objc != 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " logonHandle\"");
    return TCL_ERROR;
  }
  logonName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (AdabasLogonHandle (interp, logonName, &vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which sends a ROLLBACK request
   * to the database kernel.
   */

  return AdabasRollback (interp, vars);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaUtilCmd --
 *
 *      Implements the tcl command adautil with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid logonName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int AdaUtilCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *logonName;		/* Name of logonHandle. */
  AdabasInfo *vars;			/* logonHandle identified by 1. param. */
  char       *cmd;			/* Utility command to send (2. param). */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */

  /*
   * Check, if the first parameter is the name of a valid logonHandle.
   */
 
  if (objc != 3) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " logonHandle command\"");
    return TCL_ERROR;
  }
  logonName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (AdabasLogonHandle (interp, logonName, &vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which sends the given string as
   * utility command to the database kernel.
   */

  cmd = Tcl_GetStringFromObj (interp, objv[2], &dummyLength);
  return AdabasUtil (interp, vars, cmd);
}

/*
 *----------------------------------------------------------------------
 *
 * AdaSpecialCmd --
 *
 *      Implements the tcl command adaspecial with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid logonName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int AdaSpecialCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];       		/* Argument objects. */
{
  char       *logonName;		/* Name of logonHandle. */
  AdabasInfo *vars;			/* logonHandle identified by 1. param. */
  char        *currOption;		/* Option string (used for parsing). */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */
  int         ret;			/* Return value of send request. */
  int         thisMessType;		/* MessType (depends on 2. param) */
  long        count;			/* Value of count for switchlimit. */
  long        limit;			/* Value of limit for buflength. */
  char        data[200];		/* Buffer for switch values. */
  long        rc;			/* Numeric value of adamsg(rc). */

  /*
   * Check, if the first parameter is the name of a valid logonHandle.
   */

  if (objc < 3) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " logonHandle command [params...]\"");
    return TCL_ERROR;
  }
  logonName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (AdabasLogonHandle (interp, logonName, &vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  /*
   * Scan the second parameter as messType; dependend on this interpret
   * the following parameters.
   */

  currOption = Tcl_GetStringFromObj (interp, objv[2], &dummyLength);
  switch ((thisMessType = scanMessType (interp, currOption))) {
  case M_BUFLENGTH:
    if (objc != 4) {
      AppendResult (interp, "wrong # args: should be \"");
      AppendResult (interp,Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
      AppendResult (interp, " logonHandle buflength limit\"");
      return TCL_ERROR;
    }
    if (Tcl_GetIntFromObj (interp, objv[3], &limit) == TCL_ERROR) {
      return TCL_ERROR;
    }
    if (limit <= 0 || limit > MAXINT2) {
      sprintf (data, "limit must be between 1 and %d", MAXINT2);
      AppendResult (interp, data);
      return TCL_ERROR;
    }
    sprintf (data, "\\%03lo\\%03lo", limit/256, limit%256);
    break;

  case M_HELLO:
  case M_MINBUF:
  case M_MAXBUF:
    if (objc != 3) {
      SetResult (interp, "no parameter allowed");
      return TCL_ERROR;
    }
    strcpy (data, " ");
    break;

  case M_SWITCH:
    if (objc != 5) {
      AppendResult (interp, "wrong # args: should be \"");
      AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
      AppendResult (interp, " logonHandle switch trace debug\"");
      return TCL_ERROR;
    }
    sprintf (data, "%-20.20s%-20.20s",
	     Tcl_GetStringFromObj (interp, objv[3], &dummyLength),
	     Tcl_GetStringFromObj (interp, objv[4], &dummyLength));
    break;

  case M_SWITCHLIMIT:
    if (objc != 10) {
      AppendResult (interp, "wrong # args: should be \"");
      AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
      AppendResult (interp, " logonHandle switch trace debug start-layer");
      AppendResult (interp, " start-proc stop-layer stop-proc count\"");
      return TCL_ERROR;
    }
    if (Tcl_GetIntFromObj (interp, objv[9], &count) == TCL_ERROR) {
      return TCL_ERROR;
    }
    if (count <= 0 || count > MAXINT2) {
      sprintf (data, "count must be between 1 and %d", MAXINT2);
      AppendResult (interp, data);
      return TCL_ERROR;
    }
    sprintf (data,
	     "%-20.20s%-20.20s%-5.5s %-10.10s%-5.5s %-10.10s\\%03lo\\%03lo",
	     Tcl_GetStringFromObj (interp, objv[3], &dummyLength),
	     Tcl_GetStringFromObj (interp, objv[4], &dummyLength),
	     Tcl_GetStringFromObj (interp, objv[5], &dummyLength),
	     Tcl_GetStringFromObj (interp, objv[6], &dummyLength),
	     Tcl_GetStringFromObj (interp, objv[7], &dummyLength),
	     Tcl_GetStringFromObj (interp, objv[8], &dummyLength),
	     count/256, count%256);
    break;

  default:
    AppendResult (interp, "unknown special command \"");
    AppendResult (interp, currOption);
    AppendResult (interp, "\"");
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which sends the given string as
   * special command to the database kernel.
   */

  ret = AdabasSpecial (interp, vars, thisMessType, data);

  /*
   * Check, if the kernel returned -5 (invalid messType).
   * In this case it seems to be a quick or fast kernel (where
   * this commands are not compiled in).
   */
  if (ret == TCL_ERROR
      && AdabasGetMsgInt (interp, AdamsgRc, &rc) == TCL_OK
      && rc == -5) {
    Tcl_ResetObjResult  (interp);
    AppendResult (interp, "special command \"");
    AppendResult (interp, currOption);
    AppendResult (interp, "\" only available for a slow kernel");
  }
  return ret;
}

/*
 *----------------------------------------------------------------------
 *
 * AdaLogonCmd --
 *
 *      Implements the tcl command adalogon with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there are invalid options or
 *      the attempt to connect to the database fails for any reason,
 *      TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *
 * Side effects:
 *      There is a new logonHandle with an open database connection accessible
 *      under the returned name.
 *
 *----------------------------------------------------------------------
 */

static int
AdaLogonCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];		/* Argument Objects. */
{
  char    *connectStr;			/* String (like "krischan,geheim"). */
  char    *serverdb  = NULL;		/* Value of the -serverdb option. */
  char    *sqlMode   = NULL;		/* Value of the -sqlmode option. */
  service  service   = SQL_USER;	/* Value of the -service option. */
  ApplKind applKind  = UserAppl;	/* Scanned value of -service option. */
  long     isolation = -1;		/* Value of -isolation option as int. */
  char    *currOption;			/* Option string (used for parsing). */
  char    *currParameter;		/* Value  string (used for parsing). */
  int      length;			/* Length (used for option parsing). */
  int      c;				/* Char   (used for option parsing). */
  int      dummyLength;			/* Out parameter of GetXxxFromObj. */

  if (objc < 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " connectStr ?options?\"");
    return TCL_ERROR;
  }
  connectStr = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);

  for (objv+=2, objc-=2; objc > 1; objv += 2, objc -= 2) {
    currOption = Tcl_GetStringFromObj (interp, objv[0], &length);
    c          = currOption[1];

    if (c == 'i' && !strncmp (currOption, "-isolationlevel", length)) {
      if (Tcl_GetIntFromObj (interp, objv[1], &isolation) == TCL_ERROR) {
	return TCL_ERROR;
      }
    } else if (c == 's' && !strncmp (currOption, "-service", length)
	       && length > 5) {
      currParameter = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
      if (!strcmp (currParameter, "control")) {
	applKind = ControlAppl;
      } else if (!strcmp (currParameter, "odbc")) {
	applKind = OdbcAppl;
      } else if ((service = scanService (interp, currParameter)) < 0) {
	Tcl_ResetObjResult (interp);	
	AppendResult (interp, "Unknown service \"");
	AppendResult (interp, currParameter);
	AppendResult (interp, "\": must be user, utility, control or odbc");
	return TCL_ERROR;
      }
    } else if (c == 's' && !strncmp (currOption, "-serverdb", length)
	       && length > 5) {
      serverdb = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
    } else if (c == 's' && !strncmp (currOption, "-sqlmode", length)
	       && length > 2) {
      sqlMode = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
    } else {
      AppendResult (interp, "bad switch \"");
      AppendResult (interp, currOption);
      AppendResult (interp, "\": must be -isolationlevel, -service,");
      AppendResult (interp, " -serverdb or -sqlmode");
      return TCL_ERROR;
    }
  }

  if (objc > 0) {
    AppendResult (interp, "missing argument for switch: \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, "\"");
    return TCL_ERROR;
  }

  if (service == SQL_UTILITY && sqlMode) {
    SetResult (interp, "invalid combination of switches.");
    return TCL_ERROR;
  }

  /*
   * Now open the connection to adabas with the given parameter.
   */

  return AdabasLogon (interp, serverdb, connectStr, isolation, sqlMode,
		      service, applKind) ? TCL_OK : TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * AdaLogoffCmd --
 *
 *      Implements the tcl command adalogoff with the pre8.0 string interface.
 *
 * Results:
 *      The return value is normally TCL_OK. If there is an invalid logonName
 *      given, TCL_ERROR will be returned; in this case interp->results will
 *      be set to the error message.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int AdaLogoffCmd (dummy, interp, objc, objv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         objc;			/* Number of arguments. */
     Tcl_Obj    *objv[];		/* Argument Objects. */
{
  char       *logonName;		/* Name of logonHandle. */
  AdabasInfo *vars;			/* logonHandle identified by 1. param. */
  int         dummyLength;		/* Out parameter of GetXxxFromObj. */

  /*
   * Check, if the only parameter is the name of a valid logonHandle.
   */

  if (objc != 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, Tcl_GetStringFromObj (interp, objv[0], &dummyLength));
    AppendResult (interp, " logonHandle\"");
    return TCL_ERROR;
  }
  logonName = Tcl_GetStringFromObj (interp, objv[1], &dummyLength);
  if (AdabasLogonHandle (interp, logonName, &vars) == TCL_ERROR) {
    return TCL_ERROR;
  }

  /*
   * Call the procedure from AdabasApi, which closes the connection.
   */

  return AdabasLogoff (interp, vars);
}
