/*
 * AdabasTcl.c --
 *
 *      This module contains the main procedures for the executables
 *      adabastclsh and adabaswish, and the entry procedure for the
 *      shared library Adabastcl.
 *      It also defines the adabas command (mostly obsolete now).
 *
 * Copyright (c) 1996-1997 Christian Krone. All rights reserved.
 * This program is free software; you can redistribute it and/or
 * modify it under the same terms as Tcl itself.
 * See also licence.terms
 *
 * $Id: AdabasTcl.c,v 1.84 1997/01/12 13:21:24 adabas Exp $
 */


/*
 * Import interface (should move into AdabasTclInt.h someday)
 */

#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#ifdef ADABAS_WISH
#include <tk.h>
#endif

#include "adabas.h"
#include "sqlrte.h"
#include "AdabasPort.h"
#include "AdabasUtil.h"
#include "AdabasFormat.h"
#include "AdabasRte.h"
#include "AdabasPacket.h"
#include "AdabasSend.h"
#include "AdabasApi.h"
#include "AdabasSql.h"
#include "AdabasLong.h"

/* ----------------- Windows hack ------------------------------- */
#if defined(__WIN32__)
#   define WIN32_LEAN_AND_MEAN
#   include "windows.h"
#   undef WIN32_LEAN_AND_MEAN

/*
 * VC++ has an alternate entry point called DllMain, so we need to rename
 * our entry point.
 */

#   if defined(_MSC_VER)
#       define EXPORT(a,b) __declspec(dllexport) a b
#       define DllEntryPoint DllMain
#   else
#       if defined(__BORLANDC__)
#           define EXPORT(a,b) a _export b
#       else
#           define EXPORT(a,b) a b
#       endif
#   endif
#else
#   define EXPORT(a,b) a b
#endif

/* ----------------- Type definitions --------------------------- */

typedef struct {
  char *name;
  enum {oWithInfo=1, oMessType, oProducer, oSqlMode,
	oApplication, oMassCmd, oCommitImmediately} kind;
} OptionMap;

/* ----------------- Exported procedure headings ---------------- */
#ifdef ADABAS_MODULE
EXTERN EXPORT(int,Adabastcl_Init) _ANSI_ARGS_((Tcl_Interp *interp));
#else
extern int main _ANSI_ARGS_((int argc, char **argv));
#endif

/* ----------------- Local variables ---------------------------- */
static OptionMap sendOptions[] = {
  { "-withInfo",          oWithInfo          },
  { "-messType",          oMessType          },
  { "-producer",          oProducer          },
  { "-sqlMode",           oSqlMode           },
  { "-application",       oApplication       },
  { "-massCmd",           oMassCmd           },
  { "-commitImmediately", oCommitImmediately },
  { (char *) NULL, 0 } /* termination flag */
};

/* ----------------- Local procedure headings ------------------- */
#ifndef ADABAS_MODULE
static int Adabastcl_Init _ANSI_ARGS_((Tcl_Interp *interp));
static int static_Init    _ANSI_ARGS_((Tcl_Interp *interp));
#endif
static int AdabasCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				  int argc, char **argv));
static int AdabasOpenCmd _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp,
				      int argc, char **argv));
static int AdabasCloseCmd _ANSI_ARGS_((Tcl_Interp *interp, char *cmdName,
				       AdabasInfo *vars));
static int AdabasSendCmd _ANSI_ARGS_((ClientData clientData,
				      Tcl_Interp *interp,
				      int argc, char **argv));
static int AdabasVersion _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp,
				      int argc, char **argv));
static int AdabasXUser _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp,
				    int argc, char **argv));
static int args2xuser _ANSI_ARGS_((Tcl_Interp *interp, int argc, char **argv,
				   xuser_record *userParams));
static void xuser2result _ANSI_ARGS_((Tcl_Interp *interp,
				      xuser_record *userParams,
				      char *givenPasswd));
static int AdabasSendConfigure _ANSI_ARGS_((Tcl_Interp *interp,
					    AdabasPacketInfo *vars,
					    char *optionName));
static int scanPacketConfOpts _ANSI_ARGS_((Tcl_Interp *interp, char *expected,
					   char *optionName,
					   int argc, char **argv,
					   AdabasPacketInfo *vars));

/*
 * All the procedures, that are used to fill a packet, have the packet
 * pointer as parameter and are therefore version dependend...
 */
#if ADABAS_VERSION >= 62
static void setResultSegment _ANSI_ARGS_((Tcl_Interp *interp,
					  segment *retSegment));

#else /* ADABAS_VERSION <= 61 */
static void setResultSegment _ANSI_ARGS_((Tcl_Interp *interp,
					  packet *retSegment,
					  int sndMessType));
#endif

/* ----------------- Exported procedures ------------------------- */

#ifdef ADABAS_TCLSH
int
main (argc, argv)
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  Tcl_Main (argc, argv, Adabastcl_Init);
  return 0;
}
#endif

#ifdef ADABAS_WISH
int
main (argc, argv)
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  Tk_Main (argc, argv, Adabastcl_Init);
  return 0;
}
#endif

#ifdef ADABAS_MODULE
EXPORT(int,Adabastcl_Init)(interp)
#else
static int Adabastcl_Init (interp)
#endif
     Tcl_Interp *interp;		/* Current interpreter. */
{
  char *adabasTclDir = getenv ("ADABASTCL_DIR");
  char *dbroot       = getenv ("DBROOT");
  Tcl_DString dbRootTcl;
  char  versionString[100];
  int   versionNumber;
  Tcl_RegExp regExp;
  char *start, *end;

#if defined ADABAS_TCLSH || defined ADABAS_WISH
  if (Tcl_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
#endif

#ifdef ADABAS_WISH
  if (Tk_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
#endif

  /*
   * Experience has pointed out, that it is very important to talk to
   * a database server of the correct version; so here we check, if
   * ADABAS_VERSION is configured correctly...
   */
  AdabasRteVersion (versionString);
  regExp = Tcl_RegExpCompile (interp, "RTE *(([0-9])\\.([0-9])[0-9.]*)");
  if (Tcl_RegExpExec (interp, regExp, versionString, versionString) <= 0) {
    versionNumber = 0;
  } else {
    Tcl_RegExpRange (regExp, 2, &start, &end);
    versionNumber  = (*start-'0')*10;
    Tcl_RegExpRange (regExp, 3, &start, &end);
    versionNumber += (*start-'0');
  }
  if ((ADABAS_VERSION <= 61 && versionNumber > 61) ||
      (ADABAS_VERSION >= 62 && versionNumber < 62)) {
#if ADABAS_VERSION <= 61
    char *wanted = "upto version 6.1.1";
#else /* ADABAS_VERSION >= 62 */
    char *wanted = "of version 6.2 or above";
#endif
    char found[80];

    if (versionNumber) {
      Tcl_RegExpRange (regExp, 1, &start, &end);
      sprintf (found, "%.*s", end-start, start);
    } else {
      strcpy  (found, "unknown");
    }
    AppendResult (interp, "AdabasTcl was configured for a database ");
    AppendResult (interp, wanted);
    AppendResult (interp, ", but found a server of version ");
    AppendResult (interp, found);
    return TCL_ERROR;
  }

  AdabasSendInit (interp);

  /*
   * Here now comes the core of this procedure:
   * First create the adabas command, which is the workhorse of this extension.
   */

  Tcl_CreateCommand (interp, "adabas",
		     AdabasCmd, (ClientData) NULL,
		     (Tcl_CmdDeleteProc *) NULL);

  /*
   * Define the place, where the AdabasTcl library goes; there are the
   * sourcable .tcl files - not the shared library, which resides in
   * $DBROOT/lib, since this place you must know to execute this code).
   */

  Tcl_DStringInit (&dbRootTcl);
  if (!adabasTclDir || !*adabasTclDir) {
    if (dbroot && *dbroot) {
      Tcl_DStringAppend (&dbRootTcl, dbroot, -1);
      Tcl_DStringAppend (&dbRootTcl, "/tcl", -1);
      adabasTclDir = Tcl_DStringValue (&dbRootTcl);
    }
  }
  if (adabasTclDir) {
    Tcl_SetVar (interp, "adabastcl_library", adabasTclDir, TCL_GLOBAL_ONLY);
    Tcl_DStringFree (&dbRootTcl);
  }

  /*
   * And here are the API procedures adaopen, adasql, adafetch,...
   * (that are defined in tcl in the first version).
   */

  if (AdabasSqlInit (interp) != TCL_OK) {
    return TCL_ERROR;
  }
  AdabasApiInit (interp);

#ifdef ADABAS_TCLSH
  Tcl_SetVar(interp, "tcl_rcFileName", "~/.adabastclshrc", TCL_GLOBAL_ONLY);
#endif

#ifdef ADABAS_WISH
  Tcl_SetVar(interp, "tcl_rcFileName", "~/.adabaswishrc",  TCL_GLOBAL_ONLY);
#endif

  sprintf (versionString, "%.1f", ADABAS_TCL_VERSION);
  if (Tcl_PkgProvide (interp, "Adabastcl", versionString) == TCL_ERROR) {
    return TCL_ERROR;
  }
#ifndef ADABAS_MODULE
  Tcl_StaticPackage (interp, "Adabastcl", static_Init,
		     (Tcl_PackageInitProc *) NULL);
#endif

  return TCL_OK;
}

#ifdef ADABAS_MODULE
int
Adabastcl_SafeInit (interp)
     Tcl_Interp *interp;		/* Current interpreter. */
{
  if (Adabastcl_Init (interp) == TCL_ERROR) {
    return TCL_ERROR;
  }

  /* Here we MUST disable the adabas command before distribute this code! */

  Tcl_Eval (interp, "puts {Initialization succeeded}");
  return TCL_OK;
}
#endif

/*
 * This is some special hack just for our Unixware installation.
 */

#ifdef ADABAS_MODULE
int   sys_nerr;
char *sys_errlist[1];

/*
 * And this is a very special hack for windows...
 */

#ifdef __WIN32__
BOOL APIENTRY
DllEntryPoint(hInst,reason,reserved)
    HINSTANCE hInst;          /* Library instance handle. */
    DWORD     reason;         /* Reason this function is beeing called. */
    LPVOID    reserved;       /* Not used. */
{
    return TRUE;
}
#endif
#endif

/* ----------------- Local procedures --------------------------- */

#ifndef ADABAS_MODULE
static int
static_Init (interp)
     Tcl_Interp *interp;		/* Current interpreter. */
{
  return TCL_OK;
}
#endif

/* ------------------------------------------------------------------*/

static int
AdabasCmd (clientData, interp, argc, argv)
     ClientData  clientData;		/* Reached through to subcommands. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  size_t         length;
  int            c;

  if (argc < 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, argv[0]);
    AppendResult (interp, " option ?arg ...?\"");
    return TCL_ERROR;
  }

  c      = argv[1][0];
  length = strlen (argv[1]);

  if (c == 'o' && !strncmp (argv[1], "open", length)) {
    return AdabasOpenCmd (clientData, interp, argc-1, argv+1); 
  } else if (c == 'c' && !strncmp (argv[1], "crypt", length)) {
    return AdabasCryptCmd (clientData, interp, argc-1, argv+1); 
  } else if (c == 'g' && !strncmp (argv[1], "get", length)) {
    return AdabasGetCmd (clientData, interp, argc-1, argv+1); 
  } else if (c == 'p' && !strncmp (argv[1], "put", length)) {
    return AdabasPutCmd (clientData, interp, argc-1, argv+1); 
  } else if (c == 't' && !strncmp (argv[1], "termId", length)) {
    AdabasRteTermId (interp->result); 
    return TCL_OK;
  } else if (c == 'v' && !strncmp (argv[1], "version", length)) {
    return AdabasVersion (clientData, interp, argc-1, argv+1); 
  } else if (c == 'x' && !strncmp (argv[1], "xuser",   length)) {
    return AdabasXUser (clientData, interp, argc-1, argv+1);
  } 

  AppendResult (interp, "bad option \"");
  AppendResult (interp, argv[1]);
  AppendResult (interp, "\": should be ");
  AppendResult (interp, "open, crypt, get, put, termId, version or xuser");
  return TCL_ERROR;
}

/* ------------------------------------------------------------------*/

static int
AdabasOpenCmd (dummy, interp, argc, argv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  int         service  = SQL_USER;
  char       *serverdb = (char *) NULL;
  char       *cmdName  = "";
  int         c;
  size_t      length;
  AdabasInfo *vars;

  /*
   * Processing of the options.
   */
  for (argv+=1, argc-=1; argc > 1; argv += 2, argc -= 2) {
    c      = argv[0][1];
    length = strlen (argv[0]);

    if (c == 'n' && !strncmp (argv[0], "-name", length) && length > 1) {
      cmdName = argv[1];
    } else if (c == 's' && !strncmp (argv[0], "-service", length)
	       && length > 5) {
      if ((service = scanService (interp, argv[1])) < 0) {
	return TCL_ERROR;
      }
    } else if (c == 's' && !strncmp (argv[0], "-serverdb", length)
	       && length > 5) {
      serverdb = argv[1];
    } else {
      AppendResult (interp, "bad switch \"");
      AppendResult (interp, argv[0]);
      AppendResult (interp, "\": must be -name, -service or -serverdb");
      return TCL_ERROR;
    }
  }
  if (argc > 0) {
    AppendResult (interp, "missing argument for switch: \"");
    AppendResult (interp, argv[0]);
    AppendResult (interp, "\"");
    return TCL_ERROR;
  }

  /*
   * Initialize the database runtime environment and allocate the info var.
   */

  if (AdabasOpenConnection (interp, service, serverdb,
			    cmdName, &vars) == TCL_ERROR) {
    return TCL_ERROR;
  }
    
  /*
   * Now create the command with the evaluated name.
   */

  Tcl_CreateCommand (interp, vars->logonName,
		     AdabasSendCmd, (ClientData) vars,
		     (Tcl_CmdDeleteProc *) NULL);
  AppendResult (interp, vars->logonName);

  return TCL_OK;
}

/* ------------------------------------------------------------------*/

static int
AdabasSendCmd (clientData, interp, argc, argv)
     ClientData  clientData;		/* Logon handle (generic type). */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  AdabasInfo      *vars        = (AdabasInfo *) clientData;
  AdabasPacketInfo thisOptions = vars->packetInfo;
  int              partKind;
  size_t           length;
  int              c;
  int              partI;
  char             serverdb[DBNAME+1];

  if (argc < 2) {
    AppendResult (interp, "wrong # args: should be \"");
    AppendResult (interp, argv[0]);
    AppendResult (interp, " option ?arg ...?\"");
    return TCL_ERROR;
  }

  c      = argv[1][0];
  length = strlen (argv[1]);

  if (c == 's' && !strncmp (argv[1], "send", length)) {
    /*
     * <command> send: communicate with the database kernel by
     *                 - constructing the order packet,
     *                 - sending it to the kernel and
     *                 - waiting for the answer.
     */

    for (partI = 0; partI < PART_KINDS; partI++) {
      vars->partInfos[ partI ].bufLength = 0;
    }

#if ADABAS_VERSION <= 61
    /*
     * Since in pre6.2 there is no possibility to set autocommit
     * or withInfo, when switching the sqlmode, we must forbid this.
     */

    if (thisOptions.sqlMode != SQLM_NIL) {
      if (thisOptions.commitImmediately || thisOptions.withInfo) {
	interp->result = "No autocommit or withInfo, if sqlmode changed.";
	return TCL_ERROR;
      }
    }
#endif

    /*
     * Processing of the options.
     */

    if (argc < 3) {
      interp->result = "missing specification, what to send";
      return TCL_ERROR;      
    }
    for (argv+=2, argc-=2; argc > 1; argv += 2, argc -= 2) {
      c        = argv[0][1];
      length   = strlen (argv[0]);
      partKind = pkNil;

      if (c == 'c' && !strncmp (argv[0], "-command", length)) {
	partKind = pkCommand;
      } else if (c == 'd' && !strncmp (argv[0], "-data", length)) {
	partKind = pkData;
      } else if (c == 'l' && !strncmp (argv[0], "-longdata", length)) {
	partKind = pkLongdata;
      } else if (c == 'p' && !strncmp (argv[0], "-parsid", length)) {
	partKind = pkParsid;
      } else if (c == 'r' && !strncmp (argv[0], "-resultcount", length)) {
	partKind = pkResultcount;
      } else {
	if (scanPacketConfOpts (interp,
				"-command, -data, -resultcount, -longdata, ", 
				argv[0], argc-1, argv+1,
				&thisOptions) == TCL_ERROR) {
	  return TCL_ERROR;
	}
      }

      if (partKind != pkNil) {
	if (vars->partInfos[ partKind ].bufLength) {
	  AppendResult (interp, "\"");
	  AppendResult (interp, argv[0]);
	  AppendResult (interp, "\" is given twice");
	  return TCL_ERROR;
	}
	vars->partInfos[partKind].buf = argv[1];
	if (partKind == pkResultcount) {
	  if (!sscanf (argv[1], "%d", &vars->partInfos[partKind].bufLength)) {
	    AppendResult (interp, "Invalid result count \"");
	    AppendResult (interp, argv[1]);
	    AppendResult (interp, "\"");
	    return TCL_ERROR;
	  }
	} else {
	  vars->partInfos[partKind].bufLength = strlen (argv[1]);
	}
      }
    }

    if (argc > 0) {
      AppendResult (interp, "missing argument for switch: \"");
      AppendResult (interp, argv[0]);
      AppendResult (interp, "\"");
      return TCL_ERROR;
    }

    if (AdabasFillPacket (interp, vars, &thisOptions) == TCL_ERROR) {
      return TCL_ERROR;
    }
    if (AdabasSendPacket (interp, vars) == TCL_ERROR) {
      return TCL_ERROR;
    }

    /*
     * Transform the received packet to a valid tcl result.
     */

#if ADABAS_VERSION >= 62
    setResultSegment (interp, &vars->rcvPacket->segm);
#else /* ADABAS_VERSION <= 61 */
    setResultSegment (interp, &vars->sqlPacket, thisOptions.messType);
#endif

  } else if (c == 'c' && !strncmp (argv[1], "close", length) && length > 1) {

    /*
     * <command> close: Close a connection to the database and delete
     *                  this tcl command.
     */

    return AdabasCloseCmd (interp, argv[0], vars);

  } else if (c == 'c' && !strncmp (argv[1], "cancel", length) && length > 1) {

    /*
     * <command> cancel: Cancels the currently running sql statement.
     */

    AdabasRteCancel (&vars->rteInfo);

  } else if (c == 'r' && !strncmp (argv[1], "reopen", length)) {

    /*
     * <command> reopen: Close and then open the connection to the database.
     */

    AdabasRteClose (&vars->rteInfo);
    p2c (serverdb, vars->rteInfo.dbname, DBNAME);
    if (AdabasRteConnect (interp, &vars->rteInfo,
			  serverdb, vars->rteInfo.service) == TCL_ERROR)
      return TCL_ERROR;

  } else if (c == 'c' && !strncmp (argv[1], "configure", length)
	     && length > 1) {

    /*
     * <command> configure: Query or set the configuration.
     */

    if (argc == 2) {
      return AdabasSendConfigure (interp, &vars->packetInfo, (char *) NULL);
    } else if (argc == 3) {
      return AdabasSendConfigure (interp, &vars->packetInfo, argv[2]);
    } else {
      return scanPacketConfOpts (interp, "", argv[2], argc-3, argv+3,
				 &vars->packetInfo);
    }
  } else {
    AppendResult (interp, "bad option \"");
    AppendResult (interp, argv[1]);
    AppendResult (interp, "\": should be send, configure, cancel,");
    AppendResult (interp, " close or reopen");
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

static int
AdabasVersion (dummy, interp, argc, argv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  if (argc != 1) {
    interp->result = "wrong # args: should be adabas version";
    return TCL_ERROR;
  }
  SET_VERSION (interp->result);
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

static int
AdabasXUser (dummy, interp, argc, argv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  size_t       length;
  int          c;
  xuser_record userParams;
  int          index;
  char         givenPasswd[100]; /* more then enough. */

  if (argc == 1) {
    interp->result = "wrong # args: should be adabas xuser option ?...?";
    return TCL_ERROR;
  }
  c      = argv[1][0];
  length = strlen (argv[1]);

  if (c == 'a' && !strncmp (argv[1], "args", length)) {
    if (AdabasXUserArgs (interp, &userParams, givenPasswd) == TCL_ERROR) {
      return TCL_ERROR;
    }
    xuser2result (interp, &userParams, givenPasswd);
    return TCL_OK;
  } else if (c == 'c' && !strncmp (argv[1], "clear", length) && length > 1) {
    return AdabasXUserClear (interp);
  } else if (c == 'c' && !strncmp (argv[1], "close", length) && length > 1) {
    return AdabasXUserClose (interp);
  } else if (c == 'g' && !strncmp (argv[1], "get", length)) {
    if (argc > 3) {
      interp->result = "wrong # args: should be adabas xuser get ?userkey?";
      return TCL_ERROR;
    }
    if (AdabasXUserGet (interp, argv[2], &userParams) == TCL_ERROR) {
      return TCL_ERROR;
    }
    xuser2result (interp, &userParams, (char *) NULL);
    return TCL_OK;
  } else if (c == 'i' && !strncmp (argv[1], "index", length)) {
    if (argc != 3) {
      interp->result = "wrong # args: should be adabas xuser index no";
      return TCL_ERROR;
    }
    if (Tcl_GetInt (interp, argv[2], &index) == TCL_ERROR) {
      return TCL_ERROR;
    }
    if (AdabasXUserIndex (interp, index, &userParams) == TCL_ERROR) {
      return TCL_ERROR;
    }
    xuser2result (interp, &userParams, (char *) NULL);
    return TCL_OK;
  } else if (c == 'o' && !strncmp (argv[1], "open", length)) {
    return AdabasXUserOpen (interp);
  } else if (c == 'p' && !strncmp (argv[1], "put", length)) {
    if (argc == 2) {
      interp->result =
	"wrong # args: adabas xuser put entry value ?entry value?...";
      return TCL_ERROR;
    }
    if (args2xuser (interp, argc-2, argv+2, &userParams) == TCL_ERROR) {
      return TCL_ERROR;
    }
    return AdabasXUserPut (interp, &userParams);
  }

  AppendResult (interp, "bad option \"");
  AppendResult (interp, argv[1]);
  AppendResult (interp, "\": must be args, clear, close, get,");
  AppendResult (interp, " index, open or put");
  return TCL_ERROR;
}

/* ------------------------------------------------------------------*/

static void
xuser2result (interp, userParams, givenPasswd)
     Tcl_Interp   *interp;		/* Current interpreter. */
     xuser_record *userParams;		/* Incoming xuser record. */
     char         *givenPasswd;		/* Clear passwd from command line. */
{
  char        buffer[256];
  Tcl_DString pw;

  Tcl_AppendElement (interp, "userkey");
  p2c (buffer, userParams->xu_key,
       sizeof (userParams->xu_key));
  Tcl_AppendElement (interp, buffer);

  Tcl_AppendElement (interp, "servernode");
  p2c (buffer, userParams->xu_servernode,
       sizeof (userParams->xu_servernode));
  Tcl_AppendElement (interp, buffer);

  Tcl_AppendElement (interp, "serverdb");
  p2c (buffer, userParams->xu_serverdb,
       sizeof (userParams->xu_serverdb));
  Tcl_AppendElement (interp, buffer);

  Tcl_AppendElement (interp, "user");
  p2c (buffer, userParams->xu_user,
       sizeof (userParams->xu_user));
  Tcl_AppendElement (interp, buffer);

  if (givenPasswd && *givenPasswd) {
    Tcl_AppendElement (interp, "clearpw");
    Tcl_AppendElement (interp, givenPasswd);
  } else {
    Tcl_AppendElement (interp, "password");
    memcpy (buffer, userParams->xu_password,
	    sizeof (userParams->xu_password));
    Tcl_DStringInit (&pw);
    decodeSequence (interp, encEscape, (Tcl_Obj *) NULL, &pw, buffer, 24);
    Tcl_AppendElement (interp, Tcl_DStringValue (&pw));
    Tcl_DStringFree (&pw);
  }

  Tcl_AppendElement (interp, "sqlmode");
  p2c (buffer, userParams->xu_sqlmode,
       sizeof (userParams->xu_sqlmode));
  Tcl_AppendElement (interp, buffer);

#if ADABAS_VERSION <= 61
  Tcl_AppendElement (interp, "cachelimit");
  sprintf (buffer, "%d", userParams->xu_cachelimit);
  Tcl_AppendElement (interp, buffer);
#endif

  Tcl_AppendElement (interp, "timeout");
  sprintf (buffer, "%d", userParams->xu_timeout);
  Tcl_AppendElement (interp, buffer);

  Tcl_AppendElement (interp, "isolation");
  sprintf (buffer, "%d", userParams->xu_isolation);
  Tcl_AppendElement (interp, buffer);

#if ADABAS_VERSION >= 62
  Tcl_AppendElement (interp, "dblang");
  p2c (buffer, userParams->xu_dblang,
       sizeof (userParams->xu_dblang));
  Tcl_AppendElement (interp, buffer);
#endif
}

/* ------------------------------------------------------------------*/

static int
args2xuser (interp, argc, argv, userParams)
     Tcl_Interp   *interp;		/* Current interpreter. */
     int           argc;		/* Number of arguments. */
     char        **argv;		/* Argument strings. */
     xuser_record *userParams;		/* Returning xuser record. */
{
  size_t      length;
  int         c;
  Tcl_DString pw;
  int         value;
  int         keyFound = 0;

  /*
   * First initialize all char fields to blank and numeric fileds to -1.
   */

  memset (userParams, ' ', sizeof (*userParams));
  c2p (userParams->xu_sqlmode, sqlMode2String (SQLM_ADABAS, 1),
       sizeof (userParams->xu_sqlmode));
  userParams->xu_cachelimit = -1;
  userParams->xu_timeout    = -1;
  userParams->xu_isolation  = -1;

  /*
   * Now scan in a loop all given parameter specs.
   */

  while (argc >= 2) {
    c      = argv[0][0];
    length = strlen (argv[0]);

    switch (c) {
    case 'c':
      if (!strncmp (argv[0], "cachelimit", length)) {
	if (Tcl_GetInt (interp, argv[1], &value) == TCL_ERROR) {
	  return TCL_ERROR;
	}
	userParams->xu_cachelimit = (int4) value;
	break;
      }
#if ADABAS_VERSION >= 62
    case 'd':
      if (!strncmp (argv[0], "dblang", length)) {
	c2p (userParams->xu_dblang, argv[1],
	     sizeof (userParams->xu_dblang));
	break;
      }
#endif
    case 'i':
      if (!strncmp (argv[0], "isolation", length)) {
	if (Tcl_GetInt (interp, argv[1], &value) == TCL_ERROR) {
	  return TCL_ERROR;
	}
	userParams->xu_isolation = (int2) value;
	break;
      }
    case 's':
      if (!strncmp (argv[0], "servernode", length) && length > 6) {
	c2p (userParams->xu_servernode, argv[1],
	     sizeof (userParams->xu_servernode));
	break;
      }
      if (!strncmp (argv[0], "serverdb", length) && length > 6) {
	c2p (userParams->xu_serverdb, argv[1],
	     sizeof (userParams->xu_serverdb));
	break;
      }
      if (!strncmp (argv[0], "sqlmode", length) && length > 1) {
	if (scanSqlMode (interp, argv[1], 1) < 0) {
	  return TCL_ERROR;
	}
	c2p (userParams->xu_sqlmode, argv[1],
	     sizeof (userParams->xu_sqlmode));
	break;
      }
    case 'p':
      if (!strncmp (argv[0], "password", length)) {
	Tcl_DStringInit (&pw);
	(void) encodeSequence (interp, encEscape, "passwd",
			       &pw, argv[1], 1, 0);
	memcpy (userParams->xu_password, Tcl_DStringValue (&pw),
		sizeof (userParams->xu_password));
	Tcl_DStringFree (&pw);
	break;
      }
    case 't':
      if (!strncmp (argv[0], "timeout", length)) {
	if (Tcl_GetInt (interp, argv[1], &value) == TCL_ERROR) {
	  return TCL_ERROR;
	}
	userParams->xu_cachelimit = (int2) value;
	break;
      }
    case 'u':
      if (!strncmp (argv[0], "user", length) && length == 4) {
	c2p (userParams->xu_user, argv[1],
	     sizeof (userParams->xu_user));
	break;
      }
      if (!strncmp (argv[0], "userkey", length) && length > 4) {
	c2p (userParams->xu_key, argv[1],
	     sizeof (userParams->xu_key));
	keyFound = 1;
	break;
      }
    default:
      AppendResult (interp, "Invalid xuser entry \"");
      AppendResult (interp, argv[0]);
      AppendResult (interp, "\"");
      return TCL_ERROR;
    }
    argc -= 2;
    argv += 2;
  }

  /*
   * Is there a dangling parameter on the command line?
   */

  if (argc == 1) {
    AppendResult (interp, "Expected value for entry \"");
    AppendResult (interp, argv[0]);
    AppendResult (interp, "\"");
    return TCL_ERROR;
  }

  /*
   * One of the specifications must be the userkey itself.
   */

  if (!keyFound) {
    SetResult (interp,  "No userkey specified");
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

static int AdabasCloseCmd (interp, cmdName, vars)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *cmdName;		/* Name of logon handle to delete. */
     AdabasInfo *vars;			/* Logon handle to delete. */
{
  /*
   * Delete this tcl command.
   */

  if (Tcl_DeleteCommand (interp, cmdName) == TCL_ERROR) {
    AppendResult (interp, "can't delete command \"");
    AppendResult (interp, cmdName);
    AppendResult (interp, "\"");
    return TCL_ERROR;
  }

  return AdabasCloseConnection (interp, cmdName, vars);
}

/* ------------------------------------------------------------------*/

static int
AdabasSendConfigure (interp, vars, optionName)
     Tcl_Interp       *interp;		/* Current interpreter. */
     AdabasPacketInfo *vars;		/* Packet handle. */
     char             *optionName;	/* Name of option to examine. */
{
  int         length = optionName ? strlen (optionName) : 0;
  OptionMap  *o;
  Tcl_DString oneValue;
  char       *defaultValue;
  char       *currentValue;

  Tcl_DStringInit (&oneValue);
  for (o = sendOptions; o->name; o++) {
    if (!optionName || strncmp (optionName, o->name, length) == 0) {
      switch (o->kind) {

      case oWithInfo:
	defaultValue = "false";
	currentValue = vars->withInfo ? "true" : "false";
	break;

      case oMessType:
	defaultValue = messType2String (M_DBS);
	currentValue = messType2String (vars->messType);
	break;

      case oProducer:
#if ADABAS_VERSION >= 62
        defaultValue = producer2String (PR_USER_CMD);
	currentValue = producer2String (vars->producer);
#else /* ADABAS_VERSION <= 61 */
        defaultValue = currentValue = "user";
#endif
	break;

      case oSqlMode:
#if ADABAS_VERSION >= 62
	defaultValue = sqlMode2String (SM_SESSION_SQLMODE, 0);
#else /* ADABAS_VERSION <= 61 */
	defaultValue = sqlMode2String (SQLM_NIL, 0);
#endif
	currentValue = sqlMode2String (vars->sqlMode, 0);
	break;

      case oMassCmd:
	defaultValue = "false";
	currentValue = vars->massCmd ? "true" : "false";
	break;

      case oCommitImmediately:
	defaultValue = "false";
	currentValue = vars->commitImmediately ? "true" : "false";
	break;

      case oApplication:
	defaultValue = application2String ("XCI");
	currentValue = application2String (vars->application);
	break;

      default:
	interp->result = "Oops, invalid configure option?!?";
	return TCL_ERROR;
      }

      Tcl_DStringAppendElement (&oneValue, o->name);
      Tcl_DStringAppendElement (&oneValue, defaultValue);
      Tcl_DStringAppendElement (&oneValue, currentValue);

      if (optionName) {
	Tcl_DStringResult (interp, &oneValue);
	return TCL_OK;
      } else {
	Tcl_AppendElement (interp, Tcl_DStringValue (&oneValue));
	Tcl_DStringFree (&oneValue);
      }
    }
  }

  if (optionName) {
    AppendResult (interp, "bad switch \"");
    AppendResult (interp, optionName);
    AppendResult (interp, "\": must be -messType, -producer, -sqlMode");
    AppendResult (interp, " -massCmd, -commitImmediately or -withInfo");
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

static int
scanPacketConfOpts (interp, expected, optionName, argc, argv, vars)
     Tcl_Interp       *interp;		/* Current interpreter. */
     char             *expected;	/* List of options (for errortext). */
     char             *optionName;	/* Name of option to configure. */
     int               argc;		/* Number of arguments. */
     char            **argv;		/* Argument strings. */
     AdabasPacketInfo *vars;		/* Packet handle. */
{
  int   c      = optionName[1];
  int   length = strlen (optionName);
  int   newMessType;
  char *newApplication;
  int   newSqlMode;
#if ADABAS_VERSION >= 62
  int   newProducer;
#endif

  /*
   * <cmd> configure -application <commandInterface|load|utility...>
   * Kind of application, sending the request to the kernel.
   */
  if (c == 'a' && !strncmp (optionName, "-application", length)) {
    if (!(newApplication = scanApplication (interp, argv[0])))
      return TCL_ERROR;
    memcpy (vars->application, newApplication, 3);
    return TCL_OK;
  }

  /*
   * <cmd> configure -messType <dbs|parse|execute|utility|...>
   * Message type of packet, that will be send to the kernel.
   */

  if (c == 'm' && !strncmp (optionName, "-messType", length) && length > 2) {
    newMessType = scanMessType (interp, argv[0]);
    if (newMessType == M_NIL) return TCL_ERROR;
    vars->messType = newMessType;
    return TCL_OK;
  }

#if ADABAS_VERSION >= 62
  /*
   * <cmd> configure -producer <user|internal|installation>
   * Who is the producer of the request?
   */

  if (c == 'p' && !strncmp (optionName, "-producer", length)) {
    if ((newProducer = scanProducer (interp, argv[0])) == PR_NIL)
      return TCL_ERROR;
    vars->producer = newProducer;
    return TCL_OK;
  }
#endif

  /*
   * <cmd> configure -sqlMode <session|adabas|ansi|db2|oracle>
   * In what mode should the request be interpreted?
   */

  if (c == 's' && !strncmp (optionName, "-sqlMode", length)) {
#if ADABAS_VERSION >= 62
    if ((newSqlMode = scanSqlMode (interp, argv[0], 0)) == SM_NIL) {
      return TCL_ERROR;
    }
#else /* ADABAS_VERSION <= 61 */
    if ((newSqlMode = scanSqlMode (interp, argv[0], 0)) < 0) {
      return TCL_ERROR;
    }
#endif
    vars->sqlMode = newSqlMode;
    return TCL_OK;
  }

  /*
   * <cmd> configure -withInfo <bool>.
   * Should the kernel respond with the addition info of columnnames
   * and shortinfos for a select?
   */

  if (c == 'w' && !strncmp (optionName, "-withInfo", length)) {
    return Tcl_GetBoolean (interp, argv[0], &vars->withInfo);
  }

  /*
   * <cmd> configure -massCmd <bool>.
   * Should the given sqlStatement be executed as mass command?
   */

  if (c == 'm' && !strncmp (optionName, "-massCmd", length) && length > 2) {
    return Tcl_GetBoolean (interp, argv[0], &vars->massCmd);
  }

  /*
   * <cmd> configure -commitImmediately <bool>.
   * Should the kernel execute a commit after any ddl statement?
   */

  if (c == 'c' && !strncmp (optionName, "-commitImmediately", length)) {
    return Tcl_GetBoolean (interp, argv[0], &vars->commitImmediately);
  }

  /*
   * No known option was given, if control gets to this point.
   */

  AppendResult (interp, "bad switch \"");
  AppendResult (interp, optionName);
  AppendResult (interp, "\": must be ");
  AppendResult (interp, expected);
  AppendResult (interp, "-application, -messType, -producer,");
  AppendResult (interp, " -sqlMode, -masscmd, -commitImmediately or -withInfo");
  return TCL_ERROR;
}

/* ------------------------------------------------------------------*/

#if ADABAS_VERSION >= 62
static void setResultSegment (interp, retSegment)
     Tcl_Interp *interp;		/* Current interpreter. */
     segment    *retSegment;		/* Return segment to be converted. */
{
  part *currPart = &retSegment->part;
  int4  partLength;
  char  result[80];
  int   ix;

  sprintf (result, "-returncode %d ",
	   retSegment->segm_header.kind.ret.returncode);
  AppendResult (interp, result);
  sprintf (result, "-errorpos %d ",
	   retSegment->segm_header.kind.ret.errorpos);
  AppendResult (interp, result);
  sprintf (result, "-sqlstate %5.5s ",
	   retSegment->segm_header.kind.ret.sqlstate);
  AppendResult (interp, result);

  for (ix = 0; ix < retSegment->segm_header.kind.cmd.no_of_parts; ix++) {
    if (ix > 0) {
      AppendResult (interp, " ");
    }
    putResultPart (interp, currPart, &partLength);
    currPart = (part *) (((char *) currPart) + partLength);
  }
}
#else /* ADABAS_VERSION <= 61 */
static void setResultSegment (interp, retSegment, sndMessType)
     Tcl_Interp *interp;		/* Current interpreter. */
     packet     *retSegment;		/* Return segment to be converted. */
     int         sndMessType;		/* Mess type of sended packet. */
{
  char result[80];
  int2 currPos;

  sprintf (result, "-returncode %d ",  retSegment->return_code);
  AppendResult (interp, result);
  if (retSegment->return_code) {
    sprintf (result, "-errorpos %d ",   retSegment->error_code);
    AppendResult (interp, result);
  } else {
    /* Warning set not yet... */
  }
  currPos = 1;
  while (currPos <= retSegment->part1_length) {
    putResultPart (interp, 1, &currPos, retSegment, sndMessType);
  }
  currPos = 1;
  while (currPos <= retSegment->part2_length) {
    putResultPart (interp, 2, &currPos, retSegment, sndMessType);
  }
}
#endif
